from typing import List, Dict

import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class LongitudinalPreprocessorConfig:
    """Manages the configuration space for preprocessors

    Focus: Longitudinal supervised learning contexts

    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that will be used to add the preprocessors and their
        respective hyperparameters.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if any(
            meta_key not in config_space.meta
            for meta_key in ["data_preparation", "preprocessors"]
        ):
            raise ValueError(
                f"Expected 'data_preparation', 'preprocessors', "
                f"keys in meta of config_space. "
                f"Got {config_space.meta.keys()}"
            )
        if (
            config_space.get_hyperparameter(config_space.meta["data_preparation"])
            is None
        ):
            raise ValueError(
                f"Expected '{config_space.meta['data_preparation']}' hyperparameter "
                f"in config_space"
            )
        self.config_space = config_space
        self.preprocessors_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "CorrelationBasedFeatureSelectionPerGroup": self.setup_cfs_per_group,
            "CorrelationBasedFeatureSelection": self.setup_cfs,
            "None": self.setup_none,
        }
        self.forbidden_clauses: List[Dict[str, List[str]]] = [
            {
                "data_preparations": ["MerWavTimePlus"],
                "preprocessors_to_forbid": ["CorrelationBasedFeatureSelection"],
            },
            {
                "data_preparations": [
                    "MerWavTimeMinus",
                    "AggrFuncMean",
                    "AggrFuncMedian",
                    # "AggrFuncMode",
                    "SepWavMajorityVoting",
                    "SepWavDecayLinearWeighting",
                    "SepWavDecayExponentialWeighting",
                    "SepWavDecayCrossValidation",
                    "SepWavStackingLogisticRegression",
                    "SepWavStackingDecisionTree",
                    "SepWavStackingRandomForest",
                ],
                "preprocessors_to_forbid": ["CorrelationBasedFeatureSelectionPerGroup"],
            },
        ]
        self.cs_preprocessors_name = config_space.meta["preprocessors"]
        self.cs_data_preparation_name = config_space.meta["data_preparation"]

    @property
    def shared_hyperparameters(self):
        return {}

    def setup_preprocessors(self):
        preprocessors_choices = list(self.preprocessors_setup_map.keys())

        if not preprocessors_choices:
            raise ValueError("No preprocessors to add to config space")

        preprocessors = csh.CategoricalHyperparameter(
            name=self.cs_preprocessors_name,
            choices=preprocessors_choices,
            default_value=preprocessors_choices[0],
        )
        self.config_space.add_hyperparameter(preprocessors)

        if self.forbidden_clauses is not None:
            if (
                self.cs_data_preparation_name
                not in self.config_space.get_hyperparameter_names()
            ):
                raise ValueError(
                    f"Data preparation hyperparameter "
                    f"{self.cs_data_preparation_name} not found in config space"
                )
            data_preparation = self.config_space.get_hyperparameter(
                self.cs_data_preparation_name
            )
            data_preparation_choices = list(data_preparation.choices)
            forbidden_clauses = [
                cs.ForbiddenAndConjunction(
                    cs.ForbiddenInClause(
                        preprocessors, clause["preprocessors_to_forbid"]
                    ),
                    cs.ForbiddenInClause(data_preparation, clause["data_preparations"]),
                )
                for clause in self.forbidden_clauses
                if all(
                    data_preparation_name in data_preparation_choices
                    for data_preparation_name in clause["data_preparations"]
                )
            ]
            self.config_space.add_forbidden_clauses(forbidden_clauses)

        for preprocessor_name in preprocessors_choices:
            if setup_func := self.preprocessors_setup_map.get(preprocessor_name):
                setup_func(preprocessors)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, preprocessor_name: str
    ):
        if "preprocessors" not in local_vars or not isinstance(
            local_vars["preprocessors"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'preprocessors' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["preprocessors"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["preprocessors"], preprocessor_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_cfs(self, preprocessors: csh.CategoricalHyperparameter):
        pass

    def setup_cfs_per_group(self, preprocessors: csh.CategoricalHyperparameter):
        version = csh.CategoricalHyperparameter(
            name="version__CorrelationBasedFeatureSelectionPerGroup",
            choices=[1, 2],
            default_value=1,
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "CorrelationBasedFeatureSelectionPerGroup"
        )

    def setup_none(self, preprocessors: csh.CategoricalHyperparameter):
        pass



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class PreprocessorConfigTest:
    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "preprocessors" not in config_space.meta:
            raise ValueError("Expected 'preprocessors' key in meta of config_space")
        self.config_space = config_space
        self.preprocessors_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "SelectFwe": self.setup_select_fwe,
            "Binarizer": self.setup_binarizer,
            "FastICA": self.setup_fast_ica,
            "FeatureAgglomeration": self.setup_feature_agglomeration,
            "MaxAbsScaler": self.setup_max_abs_scaler,
            "MinMaxScaler": self.setup_min_max_scaler,
            "Normalizer": self.setup_normalizer,
            "Nystroem": self.setup_nystroem,
            "PCA": self.setup_pca,
            "PolynomialFeatures": self.setup_polynomial_features,
            "RBFSampler": self.setup_rbf_sampler,
            "RobustScaler": self.setup_robust_scaler,
            "StandardScaler": self.setup_standard_scaler,
            "SelectPercentile": self.setup_select_percentile,
            "VarianceThreshold": self.setup_variance_threshold,
        }
        self.cs_preprocessors_name = config_space.meta["preprocessors"]

    @property
    def shared_hyperparameters(self):
        return {
            "gamma": {"lower": 0.01, "upper": 1.01, "default_value": 1.0},
        }

    def setup_preprocessors(self):
        preprocessors_choices = list(self.preprocessors_setup_map.keys())

        if not preprocessors_choices:
            raise ValueError("No preprocessors to add to config space")

        preprocessors = csh.CategoricalHyperparameter(
            name=self.cs_preprocessors_name,
            choices=preprocessors_choices,
            default_value=preprocessors_choices[0],
        )
        self.config_space.add_hyperparameter(preprocessors)

        for preprocessor_name in preprocessors_choices:
            if setup_func := self.preprocessors_setup_map.get(preprocessor_name):
                setup_func(preprocessors)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, preprocessor_name: str
    ):
        if "preprocessors" not in local_vars or not isinstance(
            local_vars["preprocessors"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'preprocessors' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["preprocessors"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["preprocessors"], preprocessor_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_select_fwe(self, preprocessors: csh.CategoricalHyperparameter):
        alpha = csh.UniformFloatHyperparameter(
            "alpha__SelectFwe", 0, 0.05, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "SelectFwe")

    def setup_binarizer(self, preprocessors: csh.CategoricalHyperparameter):
        threshold = csh.UniformFloatHyperparameter(
            "threshold__Binarizer", 0.0, 1.01, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "Binarizer")

    def setup_fast_ica(self, preprocessors: csh.CategoricalHyperparameter):
        whiten = csh.CategoricalHyperparameter("whiten", ["unit-variance"])
        tol = csh.UniformFloatHyperparameter(
            "tol__FastICA", 0.0, 1.01, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "FastICA")

    def setup_feature_agglomeration(self, preprocessors: csh.CategoricalHyperparameter):
        linkage = csh.CategoricalHyperparameter(
            "linkage__FeatureAgglomeration", ["ward", "complete", "average"]
        )
        metric = csh.CategoricalHyperparameter(
            "metric__FeatureAgglomeration",
            ["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "FeatureAgglomeration"
        )

        # Forbidden clause: Linkage is different from 'ward' and metric is 'euclidean'
        forbidden_penalty_loss = cs.ForbiddenAndConjunction(
            cs.ForbiddenInClause(
                self.config_space["linkage__FeatureAgglomeration"],
                ["complete", "average"],
            ),
            cs.ForbiddenEqualsClause(
                self.config_space["metric__FeatureAgglomeration"], "euclidean"
            ),
        )
        self.config_space.add_forbidden_clause(forbidden_penalty_loss)

    def setup_max_abs_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_min_max_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_normalizer(self, preprocessors: csh.CategoricalHyperparameter):
        norm = csh.CategoricalHyperparameter("norm", ["l1", "l2", "max"])
        self._add_hyperparameters_and_equals_conditions(locals(), "Normalizer")

    def setup_nystroem(self, preprocessors: csh.CategoricalHyperparameter):
        kernel = csh.CategoricalHyperparameter(
            "kernel",
            [
                "rbf",
                "cosine",
                "chi2",
                "laplacian",
                "polynomial",
                "poly",
                "linear",
                "additive_chi2",
                "sigmoid",
            ],
        )
        gamma = csh.UniformFloatHyperparameter(
            "gamma__Nystroem", **self.shared_hyperparameters["gamma"]
        )
        n_components = csh.UniformIntegerHyperparameter("n_components", 1, 11)
        self._add_hyperparameters_and_equals_conditions(locals(), "Nystroem")

    def setup_pca(self, preprocessors: csh.CategoricalHyperparameter):
        svd_solver = csh.CategoricalHyperparameter("svd_solver", ["randomized"])
        iterated_power = csh.UniformIntegerHyperparameter("iterated_power", 1, 11)
        self._add_hyperparameters_and_equals_conditions(locals(), "PCA")

    def setup_polynomial_features(self, preprocessors: csh.CategoricalHyperparameter):
        degree = csh.CategoricalHyperparameter("degree", [2])
        include_bias = csh.CategoricalHyperparameter("include_bias", [False])
        interaction_only = csh.CategoricalHyperparameter("interaction_only", [False])
        self._add_hyperparameters_and_equals_conditions(locals(), "PolynomialFeatures")

    def setup_rbf_sampler(self, preprocessors: csh.CategoricalHyperparameter):
        gamma = csh.UniformFloatHyperparameter(
            "gamma__RBFSampler", **self.shared_hyperparameters["gamma"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "RBFSampler")

    def setup_robust_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_standard_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_select_percentile(self, preprocessors: csh.CategoricalHyperparameter):
        percentile = csh.UniformIntegerHyperparameter("percentile", 1, 100)
        self._add_hyperparameters_and_equals_conditions(locals(), "SelectPercentile")

    def setup_variance_threshold(self, preprocessors: csh.CategoricalHyperparameter):
        threshold = csh.UniformFloatHyperparameter(
            "threshold__VarianceThreshold", 0.05, 1.01, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "VarianceThreshold")



import inspect
from functools import partial
from typing import Union, Optional, List, Callable
import logging

import numpy as np
import pandas as pd
from ConfigSpace import ForbiddenEqualsClause
from gama.search_methods.base_search import BaseSearch

from gama.genetic_programming.compilers.scikitlearn import compile_individual


from gama.genetic_programming.operator_set import OperatorSet
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder
import ConfigSpace as cs

from gama.configuration.longitudinal_classification import (
    config_space as clf_config,
    LongitudinalConstraints,
    extra_hyperparameters_based_on_dataset_features_length,
)
from gama.data_loading import X_y_from_file
from gama.utilities.metrics import scoring_to_metric, Metric
from .GamaPipeline import GamaPipelineType
from .gama import Gama
from .genetic_programming.longitudinal.crossover_longitudinal import (
    random_longitudinal_crossover,
)
from .genetic_programming.longitudinal.mutation_longitudinal import (
    random_longitudinal_valid_mutation_in_place,
)
from .genetic_programming.longitudinal.operations_longitudinal import (
    create_longitudinal_random_expression,
)
from .genetic_programming.selection import eliminate_from_pareto
from .genetic_programming.longitudinal.selection_longitudinal import (
    create_longitudinal_from_population,
)
from .search_methods import AsyncEA
from .search_methods.bayesian_optimisation import BayesianOptimisation
from .utilities.config_space import get_longitudinal_estimator_by_name
from .utilities.smac import config_to_longitudinal_individual

# Avoid stopit from logging warnings every time a pipeline evaluation times out
logging.getLogger("stopit").setLevel(logging.ERROR)
log = logging.getLogger(__name__)


class GamaLongitudinalClassifier(Gama):
    """Gama with adaptations for Longitudinal classification."""

    def __init__(
        self,
        features_group: List[List[Union[int, str]]],
        non_longitudinal_features: List[Union[int, str]],
        feature_list_names: List[str],
        update_feature_groups_callback: Union[Callable, str] = "default",
        search_space: Optional[cs.ConfigurationSpace] = None,
        scoring: Metric = "roc_auc",  # type: ignore
        max_pipeline_length: Optional[int] = None,
        search: Optional[BaseSearch] = None,
        *args,
        **kwargs,
    ):
        if not search_space:
            # Do this to avoid the whole dictionary being included in the documentation.
            search_space = clf_config

        self._metrics = scoring_to_metric(scoring)

        search_space = self._search_space_check(search_space)

        if extra_hyperparameters_based_on_dataset_features_length is not None:
            for (
                extra_hyperparameter
            ) in extra_hyperparameters_based_on_dataset_features_length:
                search_space = extra_hyperparameter(search_space, feature_list_names)

        self._label_encoder = None

        if isinstance(search, BayesianOptimisation):
            search._config_to_individual_fun = config_to_longitudinal_individual
        elif isinstance(search, AsyncEA) and LongitudinalConstraints is None:
            raise ValueError(
                "LongitudinalConstraints is not available. For Longitudinal tasks, "
                "please make sure LongitudinalConstraints is available."
            )

        super().__init__(
            *args,
            search_space=search_space,
            scoring=scoring,
            search=search,
            gama_pipeline_type=GamaPipelineType.ScikitLongitudinal,
            **kwargs,
        )  # type: ignore

        self._features_group = features_group
        self._non_longitudinal_features = non_longitudinal_features
        self._update_feature_groups_callback = update_feature_groups_callback
        self._feature_list_names = feature_list_names

        if (
            "preprocessors" in self.search_space.meta
            and self.search_space.meta["preprocessors"]
            not in self.search_space.get_hyperparameter_names()
        ) or ("preprocessors" not in self.search_space.meta):
            if max_pipeline_length is None:
                log.info(
                    "Setting `max_pipeline_length` to 2 "
                    "because there are no preprocessing steps in the search space."
                )
                max_pipeline_length = 2
            elif max_pipeline_length > 2:
                raise ValueError(
                    f"`max_pipeline_length` can't be {max_pipeline_length} "
                    "because there are no preprocessing steps in the search space."
                )
        else:
            if max_pipeline_length is None:
                max_pipeline_length = 3
            elif max_pipeline_length < 3:
                log.warning(
                    f"[CAUTIOUS] Setting `max_pipeline_length` to "
                    f"{max_pipeline_length} can be problematic because there most "
                    f"probably are preprocessing steps available in the search space."
                )
            elif max_pipeline_length > 3:
                log.warning(
                    f"[CAUTIOUS] Setting `max_pipeline_length` to "
                    f"{max_pipeline_length} can be problematic because there most "
                    f"probably are only estimators/preprocessing steps available in "
                    f"the search space."
                )
        max_start_length = 3 if max_pipeline_length is None else max_pipeline_length
        self._operator_set = OperatorSet(
            mutate=partial(  # type: ignore #https://github.com/python/mypy/issues/1484
                random_longitudinal_valid_mutation_in_place,
                config_space=self.search_space,
                max_length=max_pipeline_length,
            ),
            mate=partial(random_longitudinal_crossover, max_length=max_pipeline_length),
            create_from_population=partial(
                create_longitudinal_from_population, cxpb=0.2, mutpb=0.8
            ),
            create_new=partial(
                create_longitudinal_random_expression,
                config_space=self.search_space,
                max_length=max_start_length,
            ),
            compile_=partial(
                compile_individual,
                gama_pipeline_type=self._gama_pipeline_type,
                features_group=self._features_group,
                non_longitudinal_features=self._non_longitudinal_features,
                feature_list_names=self._feature_list_names,
                update_feature_groups_callback=self._update_feature_groups_callback,
            ),
            eliminate=eliminate_from_pareto,
            evaluate_callback=self._on_evaluation_completed,
            completed_evaluations=self._evaluation_library.lookup,
            is_evaluated=self._evaluation_library.is_evaluated,
            get_search_space=lambda: self.search_space,
        )

    def _search_space_check(
        self,
        search_space: cs.ConfigurationSpace,
    ) -> cs.ConfigurationSpace:
        """Check if the search space is valid for classification."""

        # Check if the search space contains a classifier hyperparameter.
        if (
            "estimators" not in search_space.meta
            or (
                search_space.meta["estimators"]
                not in search_space.get_hyperparameters_dict()
            )
            or not isinstance(
                search_space.get_hyperparameter(search_space.meta["estimators"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the classifiers "
                "that is a CategoricalHyperparameter with choices for all desired "
                "classifiers. Please double-check the spelling of the name, and review "
                "the `meta` object in the search space configuration located at "
                "`configurations/classification.py`. The `meta` object should contain "
                "a key `estimators` with a value that is the name of the hyperparameter"
                " that contains the classifier choices."
            )

        # Check if the search space contains a preprocessor hyperparameter
        # if it is specified in the meta.
        if (
            "preprocessors" in search_space.meta
            and (
                search_space.meta["preprocessors"]
                not in search_space.get_hyperparameters_dict()
            )
            or "preprocessors" in search_space.meta
            and not isinstance(
                search_space.get_hyperparameter(search_space.meta["preprocessors"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the preprocessors "
                "that is a CategoricalHyperparameter with choices for all desired "
                "preprocessors. Please double-check the spelling of the name, and "
                "review the `meta` object in the search space configuration located at "
                "`configurations/classification.py`. The `meta` object should contain "
                "a key `preprocessors` with a value that is the name of the "
                "hyperparameter that contains the preprocessor choices. "
            )

        # Check if the search space contains a data_preparation hyperparameter
        # if it is specified in the meta.
        if (
            "data_preparation" in search_space.meta
            and (
                search_space.meta["data_preparation"]
                not in search_space.get_hyperparameters_dict()
            )
            or "data_preparation" in search_space.meta
            and not isinstance(
                search_space.get_hyperparameter(search_space.meta["data_preparation"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the data "
                "preparation that is a CategoricalHyperparameter with choices for all "
                "desired data preparations. Please double-check the spelling of the "
                "name, and review the `meta` object in the search space configuration "
                "located at `configurations/classification.py`. The `meta` object "
                "should contain a key `data_preparation` with a value that is the name"
                " of the hyperparameter that contains the data preparation choices. "
            )

        # Check if the search space contains only classifiers that have predict_proba
        # if the scoring requires probabilities.
        if any(metric.requires_probabilities for metric in self._metrics):
            # we don't want classifiers that do not have `predict_proba`,
            # because then we have to start doing one hot encodings of predictions etc.
            no_proba_clfs = []
            for classifier in search_space.get_hyperparameter(
                search_space.meta["estimators"]
            ).choices:
                if classifier == "Dummy_To_Ignore":
                    continue
                estimator = get_longitudinal_estimator_by_name(classifier)
                if (
                    estimator is not None
                    and issubclass(estimator, ClassifierMixin)
                    and not hasattr(estimator(), "predict_proba")
                ):
                    no_proba_clfs.append(classifier)

            log.info(
                f"The following classifiers do not have a predict_proba method "
                f"and will be excluded from the search space: {no_proba_clfs}"
            )
            search_space.add_forbidden_clauses(
                [
                    ForbiddenEqualsClause(
                        search_space.get_hyperparameter(
                            search_space.meta["estimators"]
                        ),
                        classifier,
                    )
                    for classifier in no_proba_clfs
                    if classifier
                ]
            )

        return search_space

    def _predict(self, x: pd.DataFrame):
        """Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array with predictions of shape (N,) where N is len(X).
        """
        y = self.model.predict(x)  # type: ignore
        # Decode the predicted labels - necessary only if ensemble is not used.
        if y[0] not in list(self._label_encoder.classes_):  # type: ignore
            y = self._label_encoder.inverse_transform(y)  # type: ignore
        return y

    def _predict_proba(self, x: pd.DataFrame):
        """Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is len(x),
             and K is the number of class labels found in `y` of `fit`.
        """
        return self.model.predict_proba(x)  # type: ignore

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        """Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            Data with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is len(x),
             and K is the number of class labels found in `y` of `fit`.
        """
        x = self._prepare_for_prediction(x)
        return self._predict_proba(x)

    def predict_proba_from_file(
        self,
        arff_file_path: str,
        target_column: Optional[str] = None,
        encoding: Optional[str] = None,
    ):
        """Predict the class probabilities for input in the arff_file.

        Parameters
        ----------
        arff_file_path: str
            An ARFF file with the same columns as the one that used in fit.
            Target column must be present in file, but its values are ignored.
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.
        encoding: str, optional
            Encoding of the ARFF file.

        Returns
        -------
        numpy.ndarray
            Numpy array with class probabilities.
            The array is of shape (N, K) where N is len(X),
            and K is the number of class labels found in `y` of `fit`.
        """
        x, _ = X_y_from_file(arff_file_path, target_column, encoding)
        x = self._prepare_for_prediction(x)
        return self._predict_proba(x)

    def fit(self, x, y, *args, **kwargs):
        """Should use base class documentation."""
        y_ = y.squeeze() if isinstance(y, pd.DataFrame) else y
        self._label_encoder = LabelEncoder().fit(y_)
        if any(isinstance(yi, str) for yi in y_):
            # If target values are `str` we encode them or scikit-learn will complain.
            y = self._label_encoder.transform(y_)
        self._evaluation_library.determine_sample_indices(stratify=y)

        # Add label information for classification to the scorer such that
        # the cross validator does not encounter unseen labels in smaller
        # data sets during pipeline evaluation.
        for m in self._metrics:
            if "labels" in inspect.signature(m.scorer._score_func).parameters:
                m.scorer._kwargs.update({"labels": y})

        super().fit(
            x=x,
            y=y,
            basic_encoding_step=False,
            fixed_encoding_step=False,
            *args,
            **kwargs,
        )

    def _encode_labels(self, y):
        self._label_encoder = LabelEncoder().fit(y)
        return self._label_encoder.transform(y)



"""
Contains mutation functions for genetic programming.
Each mutation function takes an individual and modifies it in-place.
"""
import logging
import random
from functools import partial
from typing import Callable, Optional, List, Dict, cast

import ConfigSpace as cs

from gama.genetic_programming.components import PrimitiveNode
from .constraints_longitudinal import LongitudinalSearchSpaceConstraint
from ..mutation import has_multiple_options, mut_replace_terminal
from ..components import Individual
from .operations_longitudinal import random_longitudinal_primitive_node
from ...configuration.longitudinal_classification import LongitudinalConstraints

# Avoid stopit from logging warnings every time a pipeline evaluation times out
logging.getLogger("stopit").setLevel(logging.ERROR)
log = logging.getLogger(__name__)


def mut_longitudinal_shrink(
    individual: Individual,
    _config_space: Optional[cs.ConfigurationSpace] = None,
    shrink_by: Optional[int] = None,
) -> None:
    """Mutates an Individual in-place by removing any number of Longitudinal
    primitive nodes.

    copy from gama/genetic_programming/mutation.py with some modification, which are:
    1. Shrink by is now between 1 and n_primitives - 2, instead of 1 and
    n_primitives - 1 – This is because we need at least one data preparation node
    and one estimator node
    2. We do not complete the individual by a DATA_TERMINAL, or nothing, we add at the
    end of the shrink, the data preparation node initially available in the individual
    """
    n_primitives = len(list(individual.primitives))
    if shrink_by is not None and n_primitives <= shrink_by:
        raise ValueError(f"Can't shrink size {n_primitives} individual by {shrink_by}.")
    if shrink_by is None:
        shrink_by = random.randint(1, n_primitives - 2)

    data_preparation_node = (
        LongitudinalSearchSpaceConstraint.find_data_preparation_node(individual)
    )
    current_primitive_node = individual.main_node
    primitives_left = n_primitives - 2
    while primitives_left > shrink_by:
        current_primitive_node = cast(PrimitiveNode, current_primitive_node._data_node)
        primitives_left -= 1
    current_primitive_node._data_node = data_preparation_node


def mut_replace_longitudinal_primitive(
    individual: Individual, config_space: cs.ConfigurationSpace
) -> None:
    """Mutates an Individual in-place by replacing one of its Primitives.

    copy from gama/genetic_programming/mutation.py with some modification, which are:
    1. We use LongitudinalSearchSpaceConstraint.primitive_replaceable instead of
    primitive_replaceable. This allows to take into account constraints on the
    longitudinal search space. Further information can be found in the following class:
    gama/genetic_programming/longitudinal/constraints_longitudinal.py
    2. We use random_longitudinal_primitive_node instead of random_primitive_node.
    """

    if not (
        primitives := list(
            filter(
                lambda primitive: (
                    LongitudinalSearchSpaceConstraint.primitive_replaceable(
                        primitive, config_space
                    )
                ),
                enumerate(individual.primitives),
            )
        )
    ):
        raise ValueError("Individual has no primitives suitable for replacement.")

    primitive_index, old_primitive_node = random.choice(primitives)
    exclude_options_list = LongitudinalConstraints.get_exclusions_for_node(
        individual, old_primitive_node._primitive.output
    )

    primitive_node = random_longitudinal_primitive_node(
        output_type=old_primitive_node._primitive.output,
        config_space=config_space,
        exclude=old_primitive_node._primitive,
        exclude_options_list=exclude_options_list,
    )
    individual.replace_primitive(primitive_index, primitive_node)


def mut_longitudinal_insert(
    individual: Individual, config_space: cs.ConfigurationSpace
) -> None:
    """Mutate an Individual in-place by inserting a PrimitiveNode at a random location.

    copy from gama/genetic_programming/mutation.py with some modification, which are:
    1. We use LongitudinalConstraints to get the exclusion list for the new primitive
    node. Further information can be found in the following class:
    gama/genetic_programming/longitudinal/constraints_longitudinal.py
    2. We use random_longitudinal_primitive_node instead of random_primitive_node.
    """
    parent_node = individual.main_node
    data_preparation_node = parent_node._data_node

    exclude_options_list = LongitudinalConstraints.get_exclusions_for_node(
        individual, "preprocessors"
    )

    new_primitive_node = random_longitudinal_primitive_node(
        output_type="preprocessors",
        config_space=config_space,
        exclude_options_list=exclude_options_list,
    )

    new_primitive_node._data_node = data_preparation_node
    parent_node._data_node = new_primitive_node


def random_longitudinal_valid_mutation_in_place(
    individual: Individual,
    config_space: cs.ConfigurationSpace,
    max_length: Optional[int] = None,
) -> Callable:
    """Apply a random valid mutation in place.

    copy from gama/genetic_programming/mutation.py with some modification.
    """
    n_primitives = len(list(individual.primitives))
    available_mutations: List[Callable[[Individual, Dict], None]] = []
    if max_length is not None and n_primitives > max_length:
        available_mutations.append(
            partial(mut_longitudinal_shrink, shrink_by=n_primitives - max_length)
        )
    else:
        replaceable_primitives = list(
            filter(
                lambda primitive: (
                    LongitudinalSearchSpaceConstraint.primitive_replaceable(
                        primitive, config_space
                    )
                ),
                enumerate(individual.primitives),
            )
        )

        if len(list(replaceable_primitives)) > 1:
            available_mutations.append(mut_replace_longitudinal_primitive)

        if max_length is None or n_primitives < max_length:
            available_mutations.append(mut_longitudinal_insert)
        if n_primitives > 2:
            available_mutations.append(mut_longitudinal_shrink)

        replaceable_terminals = filter(
            lambda t: has_multiple_options(
                config_space.get_hyperparameter(t.config_space_name)
            ),
            individual.terminals,
        )
        if len(list(replaceable_terminals)) > 1:
            available_mutations.append(mut_replace_terminal)

    if not available_mutations:
        log.warning(
            f"Individual {individual} has no valid mutations. "
            f"Returning original individual."
        )
        return lambda ind, config: ind

    mut_fn = random.choice(available_mutations)
    mut_fn(individual, config_space)
    return mut_fn



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class PreprocessorConfig:
    """Manages the configuration space for preprocessors in supervised learning contexts

    PreprocessorConfig oversees the configuration space of preprocessors used in
    supervised machine learning tasks. This class facilitates the addition of
    new preprocessors and the modification of existing ones in the configuration space
    via standardised methods. The ConfigSpace library is used to designate the
    configuration space, enabling the creation of complex and adaptable
    configuration setups. For additional information on using constraints and
    various types of hyperparameters with ConfigSpace, refer to
    the ConfigSpace documentation, available at:
    https://automl.github.io/ConfigSpace/main/quickstart.html

    For further details how to add, modify and remove preprocessors, refer to the
    documentation of classification task:
    /configuration/classification_task/preprocessors.py


    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that will be used to add the preprocessors and their
        respective hyperparameters.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "preprocessors" not in config_space.meta:
            raise ValueError("Expected 'preprocessors' key in meta of config_space")
        self.config_space = config_space
        self.preprocessors_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "Binarizer": self.setup_binarizer,
            "FastICA": self.setup_fast_ica,
            "FeatureAgglomeration": self.setup_feature_agglomeration,
            "MaxAbsScaler": self.setup_max_abs_scaler,
            "MinMaxScaler": self.setup_min_max_scaler,
            "Normalizer": self.setup_normalizer,
            "Nystroem": self.setup_nystroem,
            "PCA": self.setup_pca,
            "PolynomialFeatures": self.setup_polynomial_features,
            "RBFSampler": self.setup_rbf_sampler,
            "RobustScaler": self.setup_robust_scaler,
            "StandardScaler": self.setup_standard_scaler,
            "SelectFwe": self.setup_select_fwe,
            "SelectPercentile": self.setup_select_percentile,
            "VarianceThreshold": self.setup_variance_threshold,
        }

        self.cs_preprocessors_name = config_space.meta["preprocessors"]

    @property
    def shared_hyperparameters(self):
        return {
            "gamma": {"lower": 0.0, "upper": 1.01, "default_value": 0.05},
            "threshold": {"lower": 0.0, "upper": 1.01, "default_value": 0.05},
        }

    def setup_preprocessors(self):
        preprocessors_choices = list(self.preprocessors_setup_map.keys())

        if not preprocessors_choices:
            raise ValueError("No preprocessors to add to config space")

        preprocessors = csh.CategoricalHyperparameter(
            name=self.cs_preprocessors_name,
            choices=preprocessors_choices,
            default_value=preprocessors_choices[0],
        )
        self.config_space.add_hyperparameter(preprocessors)

        for preprocessor_name in preprocessors_choices:
            if setup_func := self.preprocessors_setup_map.get(preprocessor_name):
                setup_func(preprocessors)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, preprocessor_name: str
    ):
        if "preprocessors" not in local_vars or not isinstance(
            local_vars["preprocessors"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'preprocessors' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["preprocessors"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["preprocessors"], preprocessor_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_binarizer(self, preprocessors: csh.CategoricalHyperparameter):
        threshold = csh.UniformFloatHyperparameter(
            "threshold__Binarizer",
            **self.shared_hyperparameters["threshold"],
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "Binarizer")

    def setup_fast_ica(self, preprocessors: csh.CategoricalHyperparameter):
        tol = csh.UniformFloatHyperparameter(
            "tol__FastICA",
            lower=0.0,
            upper=1.01,
            default_value=0.05,
        )
        whiten = csh.CategoricalHyperparameter(
            "whiten__FastICA",
            choices=["unit-variance"],
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "FastICA")

    def setup_feature_agglomeration(self, preprocessors: csh.CategoricalHyperparameter):
        linkage = csh.CategoricalHyperparameter(
            "linkage__FeatureAgglomeration",
            choices=["ward", "complete", "average"],
        )
        metric = csh.CategoricalHyperparameter(
            "metric__FeatureAgglomeration",
            choices=["euclidean", "l1", "l2", "manhattan", "cosine"],
        )

        self._add_hyperparameters_and_equals_conditions(
            locals(), "FeatureAgglomeration"
        )

    def setup_max_abs_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_min_max_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_normalizer(self, preprocessors: csh.CategoricalHyperparameter):
        norm = csh.CategoricalHyperparameter(
            "norm__Normalizer",
            choices=["l1", "l2", "max"],
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "Normalizer")

    def setup_nystroem(self, preprocessors: csh.CategoricalHyperparameter):
        kernel = csh.CategoricalHyperparameter(
            "kernel__Nystroem",
            choices=[
                "rbf",
                "cosine",
                "chi2",
                "laplacian",
                "polynomial",
                "poly",
                "linear",
                "additive_chi2",
                "sigmoid",
            ],
        )
        gamma = csh.UniformFloatHyperparameter(
            "gamma__Nystroem",
            **self.shared_hyperparameters["gamma"],
        )
        n_components = csh.UniformIntegerHyperparameter(
            "n_components__Nystroem",
            lower=1,
            upper=11,
            default_value=1,
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "Nystroem")

    def setup_pca(self, preprocessors: csh.CategoricalHyperparameter):
        svd_solver = csh.CategoricalHyperparameter(
            "svd_solver__PCA",
            choices=["randomized"],
        )
        iterated_power = csh.UniformIntegerHyperparameter(
            "iterated_power__PCA",
            lower=1,
            upper=11,
            default_value=1,
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "PCA")

    def setup_polynomial_features(self, preprocessors: csh.CategoricalHyperparameter):
        degree = csh.CategoricalHyperparameter(
            "degree__PolynomialFeatures",
            choices=[2],
        )
        include_bias = csh.CategoricalHyperparameter(
            "include_bias__PolynomialFeatures",
            choices=[False],
        )
        interaction_only = csh.CategoricalHyperparameter(
            "interaction_only__PolynomialFeatures",
            choices=[False],
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "PolynomialFeatures")

    def setup_rbf_sampler(self, preprocessors: csh.CategoricalHyperparameter):
        gamma = csh.UniformFloatHyperparameter(
            "gamma__RBFSampler",
            **self.shared_hyperparameters["gamma"],
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "RBFSampler")

    def setup_robust_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_standard_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_select_fwe(self, preprocessors: csh.CategoricalHyperparameter):
        alpha = csh.UniformFloatHyperparameter(
            "alpha__SelectFwe",
            lower=0.0,
            upper=0.05,
            default_value=0.001,
        )
        # TODO Score func, how to add this?

        self._add_hyperparameters_and_equals_conditions(locals(), "SelectFwe")

    def setup_select_percentile(self, preprocessors: csh.CategoricalHyperparameter):
        percentile = csh.UniformIntegerHyperparameter(
            "percentile__SelectPercentile",
            lower=1,
            upper=100,
            default_value=1,
        )
        # TODO @Pieter – Score func, how to add this, you reckon?

        self._add_hyperparameters_and_equals_conditions(locals(), "SelectPercentile")

    def setup_variance_threshold(self, preprocessors: csh.CategoricalHyperparameter):
        threshold = csh.UniformFloatHyperparameter(
            "threshold__VarianceThreshold",
            lower=0.05,
            upper=1.01,
            default_value=0.05,
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "VarianceThreshold")



from typing import Union

import ConfigSpace as cs
import sklearn
from scikit_longitudinal.discovery import all_scikit_longitudinal_estimators
from scikit_longitudinal.templates import (
    CustomClassifierMixinEstimator,
    CustomTransformerMixinEstimator,
    DataPreparationMixin,
)
from sklearn.utils import all_estimators

from gama.genetic_programming.components import DATA_TERMINAL
from gama.utilities.longitudinal_config_space import (  # type: ignore
    AggrFuncMean,
    AggrFuncMedian,
    AggrFuncMode,
    SepWavMajorityVoting,
    SepWavDecayLinearWeighting,
    SepWavDecayExponentialWeighting,
    SepWavStackingLogisticRegression,
    SepWavDecayCrossValidation,
    SepWavStackingDecisionTree,
    SepWavStackingRandomForest,
    CascadeForestClassifier,
)


def get_internal_output_types() -> list[str]:
    """Returns the internal ConfigSpace/GAMA output types.

    Returns
    -------
    list[str]
        List of internal ConfigSpace/GAMA output types.
    """
    return [DATA_TERMINAL, "preprocessors", "estimators"]


def get_internal_longitudinal_output_types() -> list[str]:
    """Returns the internal ConfigSpace/GAMA output types for longitudinal tasks.

    Returns
    -------
    list[str]
        List of internal ConfigSpace/GAMA output types.
    """
    return [DATA_TERMINAL, "preprocessors", "estimators", "data_preparation"]


def get_hyperparameter_sklearn_name(hyperparameter_name: str) -> str:
    """Converts a ConfigSpace hyperparameter name to the name used in sklearn.

    Parameters
    ----------
    hyperparameter_name: str
        Name of the hyperparameter used in ConfigSpace.

    Returns
    -------
    str
        Name of the hyperparameter used in sklearn.
    """
    return hyperparameter_name.split("__")[0]


def get_estimator_by_name(name: str) -> sklearn.base.BaseEstimator:
    """Returns a (sklearn) estimator by name.

    Identify an estimator, which could be a classifier, regressor, or transformer.
    The name should be the same as the estimator's name in sklearn
    (for example, "GaussianNB"). If more than sklearn is supported, on the long term,
    this function could be improved by searching through more than sklearn.

    Parameters
    ----------
    name: str
        Name of the (sklearn) estimator.

    Returns
    -------
    estimator: sklearn.base.BaseEstimator
        The (sklearn) estimator corresponding to the name.
    """
    classifiers = dict(all_estimators(type_filter="classifier"))
    regressors = dict(all_estimators(type_filter="regressor"))
    transformers = dict(all_estimators(type_filter="transformer"))

    all_estimators_dict = classifiers | regressors | transformers

    estimator = all_estimators_dict.get(name)

    if estimator is None:
        raise ValueError(f"Could not find estimator with name {name}.")

    return estimator


def get_longitudinal_estimator_by_name(
    name: str,
) -> Union[
    CustomClassifierMixinEstimator,
    CustomTransformerMixinEstimator,
    DataPreparationMixin,
]:
    """Returns a (sklearn/scikit-longitudinal) estimator by name.

    Identify an estimator, which could be a classifier, transformer,
    longitudinal classifier, or longitudinal transformer.

    The name should be the same as the estimator's name in sklearn
    (for example, "GaussianNB"). If more than sklearn is supported, on the long term,
    this function could be improved by searching through more than sklearn.

    Parameters
    ----------
    name: str
        Name of the (sklearn) estimator.

    Returns
    -------
    estimator: sklearn.base.BaseEstimator
        The (sklearn) estimator corresponding to the name.
    """
    sklearn_classifier = dict(all_estimators(type_filter="classifier"))
    sklearn_transformer = dict(all_estimators(type_filter="transformer"))
    sklearn_longitudinal_classifier = dict(
        all_scikit_longitudinal_estimators(type_filter="classifier")
    )
    sklearn_longitudinal_transformer = dict(
        all_scikit_longitudinal_estimators(type_filter="transformer")
    )
    sklearn_longitudinal_data_preparation = dict(
        all_scikit_longitudinal_estimators(type_filter="data_preparation")
    )
    custom_longitudinal_estimators = {
        "AggrFuncMean": AggrFuncMean,
        "AggrFuncMedian": AggrFuncMedian,
        "AggrFuncMode": AggrFuncMode,
        "SepWavMajorityVoting": SepWavMajorityVoting,
        "SepWavDecayLinearWeighting": SepWavDecayLinearWeighting,
        "SepWavDecayExponentialWeighting": SepWavDecayExponentialWeighting,
        "SepWavDecayCrossValidation": SepWavDecayCrossValidation,
        "SepWavStackingLogisticRegression": SepWavStackingLogisticRegression,
        "SepWavStackingDecisionTree": SepWavStackingDecisionTree,
        "SepWavStackingRandomForest": SepWavStackingRandomForest,
    }
    custom_conventional_estimators = {
        "DeepForestClassifier": CascadeForestClassifier,
    }

    all_estimators_dict = (
        sklearn_classifier
        | sklearn_transformer
        | sklearn_longitudinal_classifier
        | sklearn_longitudinal_transformer
        | sklearn_longitudinal_data_preparation
        | custom_longitudinal_estimators
        | custom_conventional_estimators
    )

    estimator = all_estimators_dict.get(name)

    if estimator is None:
        raise ValueError(f"Could not find estimator with name {name}.")

    return estimator


def merge_configurations(
    c1: cs.ConfigurationSpace,
    c2: cs.ConfigurationSpace,
    prefix: str = "merged",
    delimiter: str = "_",
) -> cs.ConfigurationSpace:
    """Takes two configuration spaces and merges them together."""
    c1.add_configuration_space(prefix, c2, delimiter)
    return c1



""" Functions which take two Individuals and produce at least one new Individual. """
import random
from typing import List, Callable, Optional, Tuple

from gama.configuration.longitudinal_classification import LongitudinalConstraints
from gama.genetic_programming.crossover import (
    crossover_terminals,
    _shared_terminals,
    crossover_primitives,
)

from gama.genetic_programming.components import Individual


def random_longitudinal_crossover(
    ind1: Individual, ind2: Individual, max_length: Optional[int] = None
) -> Tuple[Individual, Individual]:
    """Random valid crossover between two individuals in-place, if it can be done.

    copy from gama/genetic_programming/crossover.py with some modification, which are:
    1. Now we use valid_longitudinal_crossover_functions instead of
    valid_crossover_functions
    """
    if max_length is not None and len(ind1.primitives) > max_length:
        raise ValueError(f"`individual1` ({ind1}) exceeds `max_length` ({max_length}).")
    if max_length is not None and len(ind2.primitives) > max_length:
        raise ValueError(f"`individual2` ({ind2}) exceeds `max_length` ({max_length}).")

    crossover_choices = _valid_longitudinal_crossover_functions(ind1, ind2)
    if len(crossover_choices) == 0:
        raise ValueError(f"{ind1.pipeline_str()} and {ind2.pipeline_str()} can't mate.")
    ind1, ind2 = random.choice(crossover_choices)(ind1, ind2)
    if max_length is not None and len(ind1.primitives) > max_length:
        return ind2, ind1
    return ind1, ind2


def _valid_longitudinal_crossover_functions(
    ind1: Individual, ind2: Individual
) -> List[Callable]:
    """Find all crossover functions that can produce new individuals from this input.

    copy from gama/genetic_programming/crossover.py with some modification, which are:
    1. We added one more check to see if the two longitudinal individuals can be
    crossovered, refer to the following class for further infroamtion:
    gama/genetic_programming/longitudinal/constraints_longitudinal.py
    """
    crossover_choices = []
    if list(_shared_terminals(ind1, ind2)):
        crossover_choices.append(crossover_terminals)
    if (
        len(list(ind1.primitives)) >= 2
        and len(list(ind2.primitives)) >= 2
        and LongitudinalConstraints.can_crossover(ind1, ind2)
    ):
        crossover_choices.append(crossover_primitives)
    return crossover_choices



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class RegressorConfig:
    """Manages the configuration space for regressors in supervised learning contexts

    RegressorConfig oversees the configuration space of regressors used for a
    supervised machine learning task. This class facilitates the addition of
    new regressors and the modification of existing ones in the configuration space
    via standardized methods. The ConfigSpace library is utilized to designate the
    configuration space, enabling the creation of complex and adaptable
    configuration setups. For additional information on using constraints and
    various types of hyperparameters with ConfigSpace, refer to
    the ConfigSpace documentation, available at:
    https://automl.github.io/ConfigSpace/main/quickstart.html

    For further details how to add, modify and remove regressors, refer to the
    documentation of classification task:
    /configuration/classification_task/classifiers.py

    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that defines the hyperparameters and their ranges for
        the regressors.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "estimators" not in config_space.meta:
            raise ValueError("Expected 'estimators' key in meta of config_space")
        self.config_space = config_space
        self.regressors_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "ElasticNetCV": self.setup_elastic_net_cv,
            "ExtraTreesRegressor": self.setup_extra_trees_regressor,
            "GradientBoostingRegressor": self.setup_gradient_boosting_regressor,
            "AdaBoostRegressor": self.setup_ada_boost_regressor,
            "DecisionTreeRegressor": self.setup_decision_tree_regressor,
            "KNeighborsRegressor": self.setup_k_neighbors_regressor,
            "LassoLarsCV": self.setup_lasso_lars_cv,
            "LinearSVR": self.setup_linear_svr,
            "RandomForestRegressor": self.setup_random_forest_regressor,
        }
        self.cs_estimators_name = self.config_space.meta["estimators"]

    @property
    def shared_hyperparameters(self):
        return {
            "n_estimators": [100],
            "max_features": {"lower": 0.05, "upper": 1.01, "default_value": 1.0},
            "min_samples_split": {"lower": 2, "upper": 21},
            "min_samples_leaf": {"lower": 1, "upper": 21},
            "learning_rate": [1e-3, 1e-2, 1e-1, 0.5, 1.0],
            "loss": [
                "squared_error",
                "absolute_error",
                "huber",
                "quantile",
                "linear",
                "square",
                "exponential",
            ],
            "tol": [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
            "bootstrap": [True, False],
            "max_depth": {"lower": 1, "upper": 11},
        }

    def setup_regressors(self):
        regressors_choices = list(self.regressors_setup_map.keys())

        if not regressors_choices:
            raise ValueError("No regressors to add to config space")

        regressors = csh.CategoricalHyperparameter(
            name=self.cs_estimators_name,
            choices=regressors_choices,
            default_value=regressors_choices[0],
        )
        self.config_space.add_hyperparameter(regressors)

        for regressor_name in regressors_choices:
            if setup_func := self.regressors_setup_map.get(regressor_name):
                setup_func(regressors)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, estimator_name: str
    ):
        if "regressors" not in local_vars or not isinstance(
            local_vars["regressors"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'regressors' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["regressors"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(hyperparameter, local_vars["regressors"], estimator_name)
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, classifiers: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_elastic_net_cv(self, regressors: csh.CategoricalHyperparameter):
        l1_ratio = csh.UniformFloatHyperparameter(
            "l1_ratio__ElasticNetCV", lower=0.0, upper=1.01, default_value=0.05
        )
        tol = csh.CategoricalHyperparameter(
            "tol__ElasticNetCV", self.shared_hyperparameters["tol"]
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "ElasticNetCV")

    def setup_extra_trees_regressor(self, regressors: csh.CategoricalHyperparameter):
        n_estimators = csh.Constant(
            "n_estimators__ExtraTreesRegressor",
            value=self.shared_hyperparameters["n_estimators"][0],
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__ExtraTreesRegressor",
            **self.shared_hyperparameters["max_features"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__ExtraTreesRegressor",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__ExtraTreesRegressor",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__ExtraTreesRegressor", self.shared_hyperparameters["bootstrap"]
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "ExtraTreesRegressor")

    def setup_gradient_boosting_regressor(
        self, regressors: csh.CategoricalHyperparameter
    ):
        n_estimators = csh.Constant(
            "n_estimators__GradientBoostingRegressor",
            value=self.shared_hyperparameters["n_estimators"][0],
        )
        loss = csh.CategoricalHyperparameter(
            "loss__GradientBoostingRegressor", self.shared_hyperparameters["loss"]
        )
        learning_rate = csh.CategoricalHyperparameter(
            "learning_rate__GradientBoostingRegressor",
            self.shared_hyperparameters["learning_rate"],
        )
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth__GradientBoostingRegressor",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__GradientBoostingRegressor",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__GradientBoostingRegressor",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        subsample = csh.UniformFloatHyperparameter(
            "subsample__GradientBoostingRegressor",
            lower=0.05,
            upper=1.01,
            default_value=1.0,
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__GradientBoostingRegressor",
            **self.shared_hyperparameters["max_features"],
        )
        alpha = csh.CategoricalHyperparameter(
            "alpha__GradientBoostingRegressor", [0.75, 0.8, 0.85, 0.9, 0.95, 0.99]
        )

        self._add_hyperparameters_and_equals_conditions(
            locals(), "GradientBoostingRegressor"
        )

    def setup_ada_boost_regressor(self, regressors: csh.CategoricalHyperparameter):
        n_estimators = csh.Constant(
            "n_estimators__AdaBoostRegressor",
            value=self.shared_hyperparameters["n_estimators"][0],
        )
        learning_rate = csh.CategoricalHyperparameter(
            "learning_rate__AdaBoostRegressor",
            self.shared_hyperparameters["learning_rate"],
        )
        loss = csh.CategoricalHyperparameter(
            "loss__AdaBoostRegressor", ["linear", "square", "exponential"]
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "AdaBoostRegressor")

    def setup_decision_tree_regressor(self, regressors: csh.CategoricalHyperparameter):
        max_depth = csh.UniformIntegerHyperparameter(
            "max_depth___DecisionTreeRegressor",
            **self.shared_hyperparameters["max_depth"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split___DecisionTreeRegressor",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf___DecisionTreeRegressor",
            **self.shared_hyperparameters["min_samples_leaf"],
        )

        self._add_hyperparameters_and_equals_conditions(
            locals(), "DecisionTreeRegressor"
        )

    def setup_k_neighbors_regressor(self, regressors: csh.CategoricalHyperparameter):
        n_neighbors = csh.UniformIntegerHyperparameter(
            "n_neighbors__KNeighborsRegressor", lower=1, upper=101, default_value=5
        )
        weights = csh.CategoricalHyperparameter(
            "weights__KNeighborsRegressor", ["uniform", "distance"]
        )
        p = csh.CategoricalHyperparameter("p__KNeighborsRegressor", [1, 2])

        self._add_hyperparameters_and_equals_conditions(locals(), "KNeighborsRegressor")

    def setup_lasso_lars_cv(self, regressors: csh.CategoricalHyperparameter):
        self._add_hyperparameters_and_equals_conditions(locals(), "LassoLarsCV")

    def setup_linear_svr(self, regressors: csh.CategoricalHyperparameter):
        loss = csh.CategoricalHyperparameter(
            "loss__LinearSVR", ["epsilon_insensitive", "squared_epsilon_insensitive"]
        )
        dual = csh.CategoricalHyperparameter("dual__LinearSVR", [True, False])
        tol = csh.CategoricalHyperparameter(
            "tol__LinearSVR", self.shared_hyperparameters["tol"]
        )
        C = csh.CategoricalHyperparameter(
            "C__LinearSVR",
            [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        )
        epsilon = csh.CategoricalHyperparameter(
            "epsilon__LinearSVR", [1e-4, 1e-3, 1e-2, 1e-1, 1.0]
        )

        self._add_hyperparameters_and_equals_conditions(locals(), "LinearSVR")

    def setup_random_forest_regressor(self, regressors: csh.CategoricalHyperparameter):
        n_estimators = csh.Constant(
            "n_estimators__RandomForestRegressor",
            value=self.shared_hyperparameters["n_estimators"][0],
        )
        max_features = csh.UniformFloatHyperparameter(
            "max_features__RandomForestRegressor",
            **self.shared_hyperparameters["max_features"],
        )
        min_samples_split = csh.UniformIntegerHyperparameter(
            "min_samples_split__RandomForestRegressor",
            **self.shared_hyperparameters["min_samples_split"],
        )
        min_samples_leaf = csh.UniformIntegerHyperparameter(
            "min_samples_leaf__RandomForestRegressor",
            **self.shared_hyperparameters["min_samples_leaf"],
        )
        bootstrap = csh.CategoricalHyperparameter(
            "bootstrap__RandomForestRegressor",
            self.shared_hyperparameters["bootstrap"],
        )

        self._add_hyperparameters_and_equals_conditions(
            locals(), "RandomForestRegressor"
        )



import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh


class PreprocessorConfig:
    """Manages the configuration space for preprocessors in supervised learning contexts

    PreprocessorConfig oversees the configuration space of preprocessors used in
    supervised machine learning tasks. This class facilitates the addition of
    new preprocessors and the modification of existing ones in the configuration space
    via standardised methods. The ConfigSpace library is used to designate the
    configuration space, enabling the creation of complex and adaptable
    configuration setups. For additional information on using constraints and
    various types of hyperparameters with ConfigSpace, refer to
    the ConfigSpace documentation, available at:
    https://automl.github.io/ConfigSpace/main/quickstart.html

    Add a preprocessor 💡
    ----------------

    1️⃣ To add a new preprocessor, define its setup method following the naming
    convention `setup_preprocessorName`. This method should:
        * Define hyperparameters specific to the preprocessor.
        * Use `_add_hyperparameters_and_equals_conditions` to add these
        hyperparameters to the config space with appropriate conditions.

    2️⃣ Next, your setup method needs to be added to the `preprocessors_setup_map` in
    the `__init__` method, where the key should be the Sci-kit learn name of your
    preprocessor, and the value should be pointing to your newly setup method.

    voila! 🎉 You are done! Your preprocessor is now added to the config space.

    How to use the shared hyperparameters 🪢
    -------------------------------------

    The shared hyperparameters are hyperparameters that are shared across multiple
    preprocessors. These hyperparameters are defined in the `shared_hyperparameters`
    property. To use these hyperparameters, simply add them to the setup method of
    the preprocessor you are adding. For example, to add the `gamma` hyperparameter to
    the `Nystroem` preprocessor, add the following line to the `setup_nystroem` method:

    >>>    gamma = csh.CategoricalHyperparameter(
    >>>        "gamma__Nystroem", self.shared_hyperparameters["gamma"]
    >>>    )

    voila! 🎉 The `gamma` hyperparameter is now added to the Nystroem preprocessor
    with the shared value available in the `shared_hyperparameters` property.

        How to name my hyperparameters ✍️
    ------------------------------

    The hyperparameters you add to the config space should be named in the following
    format if similar hyperparameter names can be found in other preprocessors:

    >>>    <hyperparameter_name>__<PreprocessorName>

    For example, the `gamma` hyperparameter for the `Nystroem` preprocessor should
    be named `gamma__Nystroem` given that the `gamma` hyperparameter is also
    available in the `RBFSampler` preprocessor. This naming convention is used to ensure
    that the hyperparameters are added to the correct preprocessor in the config space.

    If your hyperparameter name is unique to your preprocessor, you can name it as you
    please without the need to have `__<PreprocessorName>` at the end of the name.
    Nonetheless, following the naming convention would in any way not cause any issues.

    Modify an existing preprocessor 💅
    -------------------

    To modify an existing preprocessor, adjust its respective setup method and the
    shared hyperparameters property as needed by modifying the values of the
    hyperparameters. For example, to change the value of the `gamma` hyperparameter for
    the `Nystroem` preprocessor, change the value of the `gamma` hyperparameter
    in the `shared_hyperparameters` property by:

    >>>    "gamma": {"lower": 0.001, "upper": 0.8, "default_value": 0.5},

    The `gamma` hyperparameter will then be added to the config space with the
    appropriate value. However, be cautious, if you change values in the shared
    hyperparameters property, it will be changed for all preprocessors that use that
    hyperparameter. If you want this change to only apply to a specific preprocessor,
    you should add the hyperparameter to the setup method of that preprocessor.
    E.g. if you want to change the value of the `gamma` hyperparameter for the
    `Nystroem` preprocessor, and only want this change to apply to the `Nystroem`
    preprocessor, add the following line to the `setup_nystroem` method:

    >>>    gamma = csh.CategoricalHyperparameter(
    >>>        "gamma__Nystroem", {"lower": 0.001, "upper": 0.8, "default_value": 0.5},
    >>>    )

    The `gamma` hyperparameter will be added as-is for the `Nystroem` preprocessor
    and the value of the `gamma` hyperparameter for other preprocessors will be as
    available in the `shared_hyperparameters` property – iff they use the `gamma`
    hyperparameter of the `shared_hyperparameters` property.


    Parameters
    ----------
    config_space : cs.ConfigurationSpace
        The ConfigSpace object that will be used to add the preprocessors and their
        respective hyperparameters.

    """

    def __init__(
        self,
        config_space: cs.ConfigurationSpace,
    ):
        if "preprocessors" not in config_space.meta:
            raise ValueError("Expected 'preprocessors' key in meta of config_space")
        self.config_space = config_space
        self.preprocessors_setup_map = {
            "Dummy_To_Ignore": self.setup_dummy_to_ignore,
            "SelectFwe": self.setup_select_fwe,
            "Binarizer": self.setup_binarizer,
            "FastICA": self.setup_fast_ica,
            "FeatureAgglomeration": self.setup_feature_agglomeration,
            "MaxAbsScaler": self.setup_max_abs_scaler,
            "MinMaxScaler": self.setup_min_max_scaler,
            "Normalizer": self.setup_normalizer,
            "Nystroem": self.setup_nystroem,
            "PCA": self.setup_pca,
            "PolynomialFeatures": self.setup_polynomial_features,
            "RBFSampler": self.setup_rbf_sampler,
            "RobustScaler": self.setup_robust_scaler,
            "StandardScaler": self.setup_standard_scaler,
            "SelectPercentile": self.setup_select_percentile,
            "VarianceThreshold": self.setup_variance_threshold,
        }
        self.cs_preprocessors_name = config_space.meta["preprocessors"]

    @property
    def shared_hyperparameters(self):
        return {
            "gamma": {"lower": 0.01, "upper": 1.01, "default_value": 1.0},
        }

    def setup_preprocessors(self):
        preprocessors_choices = list(self.preprocessors_setup_map.keys())

        if not preprocessors_choices:
            raise ValueError("No preprocessors to add to config space")

        preprocessors = csh.CategoricalHyperparameter(
            name=self.cs_preprocessors_name,
            choices=preprocessors_choices,
            default_value=preprocessors_choices[0],
        )
        self.config_space.add_hyperparameter(preprocessors)

        for preprocessor_name in preprocessors_choices:
            if setup_func := self.preprocessors_setup_map.get(preprocessor_name):
                setup_func(preprocessors)

    def _add_hyperparameters_and_equals_conditions(
        self, local_vars: dict, preprocessor_name: str
    ):
        if "preprocessors" not in local_vars or not isinstance(
            local_vars["preprocessors"], csh.CategoricalHyperparameter
        ):
            raise ValueError(
                "Expected 'preprocessors' key with a CategoricalHyperparameter in local"
                "vars"
            )

        hyperparameters_to_add = [
            hyperparameter
            for hyperparameter in local_vars.values()
            if isinstance(hyperparameter, csh.Hyperparameter)
            and hyperparameter != local_vars["preprocessors"]
        ]

        conditions_to_add = [
            cs.EqualsCondition(
                hyperparameter, local_vars["preprocessors"], preprocessor_name
            )
            for hyperparameter in hyperparameters_to_add
        ]

        self.config_space.add_hyperparameters(hyperparameters_to_add)
        self.config_space.add_conditions(conditions_to_add)

    def setup_dummy_to_ignore(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_select_fwe(self, preprocessors: csh.CategoricalHyperparameter):
        alpha = csh.UniformFloatHyperparameter(
            "alpha__SelectFwe", 0.01, 0.05, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "SelectFwe")

    def setup_binarizer(self, preprocessors: csh.CategoricalHyperparameter):
        threshold = csh.UniformFloatHyperparameter(
            "threshold__Binarizer", 0.0, 1.01, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "Binarizer")

    def setup_fast_ica(self, preprocessors: csh.CategoricalHyperparameter):
        whiten = csh.CategoricalHyperparameter("whiten", ["unit-variance"])
        tol = csh.UniformFloatHyperparameter(
            "tol__FastICA", 0.0, 1.01, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "FastICA")

    def setup_feature_agglomeration(self, preprocessors: csh.CategoricalHyperparameter):
        linkage = csh.CategoricalHyperparameter(
            "linkage__FeatureAgglomeration", ["ward", "complete", "average"]
        )
        metric = csh.CategoricalHyperparameter(
            "metric__FeatureAgglomeration",
            ["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"],
        )
        self._add_hyperparameters_and_equals_conditions(
            locals(), "FeatureAgglomeration"
        )

        # Forbidden clause: Linkage is different from 'ward' and metric is 'euclidean'
        forbidden_penalty_loss = cs.ForbiddenAndConjunction(
            cs.ForbiddenInClause(
                self.config_space["linkage__FeatureAgglomeration"],
                ["complete", "average"],
            ),
            cs.ForbiddenEqualsClause(
                self.config_space["metric__FeatureAgglomeration"], "euclidean"
            ),
        )
        self.config_space.add_forbidden_clause(forbidden_penalty_loss)

    def setup_max_abs_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_min_max_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_normalizer(self, preprocessors: csh.CategoricalHyperparameter):
        norm = csh.CategoricalHyperparameter("norm", ["l1", "l2", "max"])
        self._add_hyperparameters_and_equals_conditions(locals(), "Normalizer")

    def setup_nystroem(self, preprocessors: csh.CategoricalHyperparameter):
        kernel = csh.CategoricalHyperparameter(
            "kernel",
            [
                "rbf",
                "cosine",
                "chi2",
                "laplacian",
                "polynomial",
                "poly",
                "linear",
                "additive_chi2",
                "sigmoid",
            ],
        )
        gamma = csh.UniformFloatHyperparameter(
            "gamma__Nystroem", **self.shared_hyperparameters["gamma"]
        )
        n_components = csh.UniformIntegerHyperparameter("n_components", 1, 11)
        self._add_hyperparameters_and_equals_conditions(locals(), "Nystroem")

    def setup_pca(self, preprocessors: csh.CategoricalHyperparameter):
        svd_solver = csh.CategoricalHyperparameter("svd_solver", ["randomized"])
        iterated_power = csh.UniformIntegerHyperparameter("iterated_power", 1, 11)
        self._add_hyperparameters_and_equals_conditions(locals(), "PCA")

    def setup_polynomial_features(self, preprocessors: csh.CategoricalHyperparameter):
        degree = csh.CategoricalHyperparameter("degree", [2])
        include_bias = csh.CategoricalHyperparameter("include_bias", [False])
        interaction_only = csh.CategoricalHyperparameter("interaction_only", [False])
        self._add_hyperparameters_and_equals_conditions(locals(), "PolynomialFeatures")

    def setup_rbf_sampler(self, preprocessors: csh.CategoricalHyperparameter):
        gamma = csh.UniformFloatHyperparameter(
            "gamma__RBFSampler", **self.shared_hyperparameters["gamma"]
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "RBFSampler")

    def setup_robust_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_standard_scaler(self, preprocessors: csh.CategoricalHyperparameter):
        # No hyperparameters
        pass

    def setup_select_percentile(self, preprocessors: csh.CategoricalHyperparameter):
        percentile = csh.UniformIntegerHyperparameter("percentile", 1, 100)
        self._add_hyperparameters_and_equals_conditions(locals(), "SelectPercentile")

    def setup_variance_threshold(self, preprocessors: csh.CategoricalHyperparameter):
        threshold = csh.UniformFloatHyperparameter(
            "threshold__VarianceThreshold", 0.05, 1.01, default_value=0.05
        )
        self._add_hyperparameters_and_equals_conditions(locals(), "VarianceThreshold")



""" Selection operators. """
import random
from typing import List

from gama.genetic_programming.components import Individual
from .crossover_longitudinal import _valid_longitudinal_crossover_functions
from gama.genetic_programming.nsga2 import nsga2_select
from gama.genetic_programming.operator_set import OperatorSet


def create_longitudinal_from_population(
    operator_shell: OperatorSet,
    pop: List[Individual],
    n: int,
    cxpb: float,
    mutpb: float,
) -> List[Individual]:
    """Creates n new individuals based on the population.

    copy from gama/genetic_programming/selection.py with some modification, which are:
    1. We now use valid_longitudinal_crossover_functions instead of
    valid_crossover_functions.
    """
    offspring = []
    metrics = [lambda ind: ind.fitness.values[0], lambda ind: ind.fitness.values[1]]
    parent_pairs = nsga2_select(pop, n, metrics)
    for ind1, ind2 in parent_pairs:
        if (
            random.random() < cxpb
            and len(_valid_longitudinal_crossover_functions(ind1, ind2)) > 0
        ):
            ind1 = operator_shell.mate(ind1, ind2)
        else:
            ind1 = operator_shell.mutate(ind1)
        offspring.append(ind1)
    return offspring



import logging
from asyncio import Future
from typing import List, Optional, Tuple, Callable, Any

import numpy as np
import pandas as pd
from smac.facade import AbstractFacade
from smac.runhistory import StatusType
from smac.runhistory.dataclasses import TrialValue, TrialInfo

from gama.genetic_programming.components import Individual
from gama.genetic_programming.operator_set import OperatorSet
from gama.search_methods.base_search import (
    BaseSearch,
    _check_base_search_hyperparameters,
)
from gama.utilities.generic.async_evaluator import AsyncEvaluator
from gama.utilities.smac import (
    get_smac,
    validate_future_valid,
    config_to_individual,
)
import ConfigSpace as cs

log = logging.getLogger(__name__)


class BayesianOptimisation(BaseSearch):
    """Perform Bayesian Optimisation over all possible pipelines."""

    def __init__(
        self,
        scenario_params: Optional[dict] = None,
        initial_design_params: Optional[dict] = None,
        facade_params: Optional[dict] = None,
        config_to_individual_fun: Callable = config_to_individual,
        **kwargs,
    ):
        super().__init__()
        self._scenario_params = scenario_params
        self._initial_design_params = initial_design_params
        self._facade_params = facade_params
        self._smac: Optional[AbstractFacade] = None
        self._config_to_individual_fun = config_to_individual_fun
        self._output_directory = None
        self.random_state = None

    def dynamic_defaults(
        self, x: pd.DataFrame, y: pd.DataFrame, time_limit: float
    ) -> None:
        def set_gama_smac_default_scenario(gama_param: str, smac_param: Any) -> None:
            if (
                self._scenario_params
                and gama_param not in self._scenario_params
                and smac_param
            ):
                self._scenario_params[gama_param] = smac_param
            elif not self._scenario_params and smac_param:
                self._scenario_params = {gama_param: smac_param}

        for gama_param, smac_param in {
            "output_directory": self._output_directory,
            "seed": self.random_state,
        }.items():
            set_gama_smac_default_scenario(gama_param, smac_param)

    def search(
        self, operations: OperatorSet, start_candidates: List[Individual]
    ) -> None:
        if (
            operations.get_search_space is None
            or (config_space := operations.get_search_space()) is None
        ):
            raise ValueError("BayesianOptimisation: Config space is None in 'search'.")
        self.smac = get_smac(
            configSpace=config_space,
            scenario_params=self._scenario_params,
            initial_design_params=self._initial_design_params,
            facade_params=self._facade_params,
        )

        self.output, self.smac = bayesian_optimisation(
            config_to_individual_fun=self._config_to_individual_fun,
            operations=operations,
            output=self.output,
            start_candidates=start_candidates,
            smac=self.smac,
        )


def bayesian_optimisation(
    config_to_individual_fun: Callable,
    operations: OperatorSet,
    output: List[Individual],
    start_candidates: List[Individual],
    smac: AbstractFacade,
    max_evaluations: Optional[int] = None,
    max_attempts: int = 100000,
) -> Tuple[List[Individual], AbstractFacade]:
    """Perform Bayesian Optimisation over all possible pipelines."""

    _check_base_search_hyperparameters(operations, output, start_candidates)
    with AsyncEvaluator() as async_:

        def smac_ask_and_submit() -> TrialInfo:
            """Ask SMAC for a configuration to turn into an individual for evaluation."""

            def _ask() -> TrialInfo:
                """Ask SMAC for a configuration"""
                if (
                    smac is None
                    or not (info_cand := smac.ask())
                    or info_cand.seed is None
                ):
                    raise ValueError(
                        "BayesianOptimisation: SMAC ask failed in 'smac_ask_and_submit'. "
                        "SMAC object or smac.ask().seed should not be None."
                    )
                return info_cand

            def ask_and_ignore_dummy_techniques(
                technique_name_to_ignore: str = "Dummy_To_Ignore",
            ) -> cs.Configuration:
                """ConfigSpace do not allow adding forbidden clauses on the default choice of a hyperparameter.

                Dummy_To_Ignore becomes default and is ignored / found an alternative when picked by sample_configuration.
                Nonetheless, for better maintainability, "Dummy_To_Ignore" can be changed, for example in the future it could be
                in the meta of the config so that it can be changed in one place by the user, i.e in /configuratins.
                """
                new_candidate = _ask()
                config = new_candidate.config

                is_preprocessor_or_estimator_to_ignore = (
                    technique_name_to_ignore in config.values()
                )
                while is_preprocessor_or_estimator_to_ignore:
                    temp_candidate = _ask()
                    temp_config = temp_candidate.config

                    is_preprocessor_or_estimator_to_ignore = (
                        technique_name_to_ignore in temp_config.values()
                    )
                    if not is_preprocessor_or_estimator_to_ignore:
                        return temp_candidate
                return new_candidate

            candidate = ask_and_ignore_dummy_techniques()
            attempts = 0
            while True:
                if not (
                    individual := config_to_individual_fun(candidate.config, operations)
                ):
                    raise ValueError(
                        "BayesianOptimisation: Conversion of SMAC config to GAMA individual"
                        "failed in 'smac_ask_and_submit'."
                    )
                if operations.is_evaluated is None:
                    raise ValueError(
                        "BayesianOptimisation: Operations.is_evaluated is None in "
                        "'smac_ask_and_submit'."
                    )
                if not operations.is_evaluated(individual):
                    async_.submit(operations.evaluate, individual)
                    break

                attempts += 1
                if attempts >= max_attempts:
                    raise ValueError(
                        "Maximum attempts reached while trying to generate a"
                        "unique individual."
                    )
                candidate = ask_and_ignore_dummy_techniques()

            return candidate

        @validate_future_valid
        def smac_handle_and_tell(future: Future, info: TrialInfo) -> None:
            """Handle the result of an evaluation and update SMAC with it."""
            if smac is None:
                raise ValueError("BayesianOptimisation: SMAC object is None.")

            individual = future.result.individual  # type: ignore

            if (
                fitness_values := individual.fitness.values
            ) and np.inf in fitness_values:
                log.warning(
                    f"BayesianOptimisation: The pipeline crashed during evaluation. "
                    f"The cost is set to -1. Individual: {individual}"
                    f"Fitness Values: {fitness_values}"
                )
                cost = -1
            else:
                cost = 1 - fitness_values[0]

            start_time = individual.fitness.start_time.timestamp()
            trial_value = TrialValue(
                cost=cost,
                time=individual.fitness.wallclock_time,
                status=(
                    StatusType.CRASHED
                    if hasattr(individual.fitness, "error")
                    else StatusType.SUCCESS
                ),
                starttime=start_time,
                endtime=start_time + individual.fitness.process_time,
            )

            output.append(individual)
            smac.tell(info, trial_value)

        if start_candidates:
            log.warning(
                "BayesianOptimisation: No start candidates are evaluated as of the "
                "current version. Further will be implemented to "
                "support both start candidates and SMAC warms-tart using GAMA meta"
                "learning, in future versions."
            )

        info = smac_ask_and_submit()

        while (max_evaluations is None) or (len(output) < max_evaluations):
            if (future := operations.wait_next(async_)).result is not None:
                smac_handle_and_tell(future, info)
            info = smac_ask_and_submit()

    return output, smac



import pandas as pd

from .gama import Gama
from gama.configuration.regression import config_space as reg_config
import ConfigSpace as cs


class GamaRegressor(Gama):
    """Gama with adaptations for regression."""

    def __init__(
        self, search_space=None, scoring="neg_mean_squared_error", *args, **kwargs
    ):
        """ """
        # Empty docstring overwrites base __init__ doc string.
        # Prevents duplication of the __init__ doc string on the API page.

        if not search_space:
            search_space = reg_config

        search_space = self._search_space_check(search_space)

        super().__init__(*args, search_space=search_space, scoring=scoring, **kwargs)

    def _search_space_check(
        self, search_space: cs.ConfigurationSpace
    ) -> cs.ConfigurationSpace:
        """Check if the search space is valid for regression."""

        # Check if the search space contains a regressor hyperparameter.
        if (
            "estimators" not in search_space.meta
            or (
                search_space.meta["estimators"]
                not in search_space.get_hyperparameters_dict()
            )
            or not isinstance(
                search_space.get_hyperparameter(search_space.meta["estimators"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the regressors "
                "that is a CategoricalHyperparameter with choices for all desired "
                "regressors. Please double-check the spelling of the name, and review "
                "the `meta` object in the search space configuration located at "
                "`configurations/regression.py`. The `meta` object should contain "
                "a key `estimators` with a value that is the name of the hyperparameter"
                " that contains the regressor choices."
            )

        # Check if the search space contains a preprocessor hyperparameter
        # if it is specified in the meta.
        if (
            "preprocessors" in search_space.meta
            and (
                search_space.meta["preprocessors"]
                not in search_space.get_hyperparameters_dict()
            )
            or "preprocessors" in search_space.meta
            and not isinstance(
                search_space.get_hyperparameter(search_space.meta["preprocessors"]),
                cs.CategoricalHyperparameter,
            )
        ):
            raise ValueError(
                "The search space must include a hyperparameter for the preprocessors "
                "that is a CategoricalHyperparameter with choices for all desired "
                "preprocessors. Please double-check the spelling of the name, and "
                "review the `meta` object in the search space configuration located at "
                "`configurations/regression.py`. The `meta` object should contain "
                "a key `preprocessors` with a value that is the name of the "
                "hyperparameter that contains the preprocessor choices. "
            )

        return search_space

    def _predict(self, x: pd.DataFrame):
        """Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe the same number of columns as that of X of `fit`.

        Returns
        -------
        numpy.ndarray
            Array with predictions of shape (N,) where N is len(X).
        """
        return self.model.predict(x)  # type: ignore



from math import sqrt, log2
from typing import List, Union

import ConfigSpace as cs
import ConfigSpace.hyperparameters as csh

from gama.configuration.longitudinal_classification_task.classifiers import (
    LongitudinalClassifierConfig,
)
from gama.configuration.longitudinal_classification_task.data_preparation import (
    LongitudinalDataPreparationConfig,
)
from gama.configuration.longitudinal_classification_task.preprocessors import (
    LongitudinalPreprocessorConfig,
)
from gama.genetic_programming.longitudinal.constraints_longitudinal import (
    LongitudinalSearchSpaceConstraint,
)

# Classifiers & Preprocessors 🚀

LongitudinalConstraints = LongitudinalSearchSpaceConstraint(
    constraints={
        "MerWavTimePlus": {
            "data_preparation": ["MerWavTimePlus"],
            "preprocessors": ["CorrelationBasedFeatureSelection", "None"],
            "estimators": [
                "DecisionTreeClassifier",
                "RandomForestClassifier",
                "ExtraTreesClassifier",
                "KNeighborsClassifier",
                "LinearSVC",
                "DeepForestClassifier",
                "GradientBoostingClassifier",
            ],
        },
        "Default": {
            "data_preparation": ["MerWavTimePlus"],
            "preprocessors": ["CorrelationBasedFeatureSelectionPerGroup", "None"],
            "estimators": [
                "NestedTreesClassifier",
                "LexicoRandomForestClassifier",
                "LexicoDecisionTreeClassifier",
                "LexicoDeepForestClassifier",
                "LexicoGradientBoostingClassifier",
            ],
        },
    }
)

config_space = cs.ConfigurationSpace(
    meta={
        # "gama_system_name": "current_configuration_name",
        "data_preparation": "data_preparation",
        "preprocessors": "preprocessors",
        "estimators": "classifiers",
    }
)

data_preparation_config = LongitudinalDataPreparationConfig(config_space)
data_preparation_config.setup_data_preparation()

classifier_config = LongitudinalClassifierConfig(config_space)
classifier_config.setup_classifiers()

preprocessor_config = LongitudinalPreprocessorConfig(config_space)
preprocessor_config.setup_preprocessors()


# Extra Dynamic Hyperparameters 🚀


def max_features_based_on_dataset_features_length(
    search_space: cs.ConfigurationSpace, features: List[str], algorithm_name: str
) -> cs.ConfigurationSpace:
    def compute_max_features_float(value: Union[int, float]) -> Union[int, float]:
        return value / len(features)

    # Check if the CategoryHyperparameter already exists
    # if so do not add any new hyperparameters
    if f"max_features__{algorithm_name}" in search_space.get_hyperparameters_dict():
        return search_space

    max_features = csh.CategoricalHyperparameter(
        name=f"max_features__{algorithm_name}",
        choices=[
            compute_max_features_float(sqrt(len(features))),
            compute_max_features_float(sqrt(len(features)) * 2),
            compute_max_features_float(sqrt(len(features)) / 2),
            compute_max_features_float(log2(len(features)) + 1),
        ],
    )
    search_space.add_hyperparameter(max_features)
    search_space.add_condition(
        cs.EqualsCondition(
            max_features,
            search_space.get_hyperparameter("classifiers"),
            algorithm_name,
        )
    )
    return search_space


extra_hyperparameters_based_on_dataset_features_length = [
    lambda search_space, features, algorithm_name=algorithm_name: max_features_based_on_dataset_features_length(  # noqa: E501
        search_space, features, algorithm_name
    )
    for algorithm_name in [
        "LexicoRandomForestClassifier",
        "RandomForestClassifier",
        "ExtraTreesClassifier",
        "GradientBoostingClassifier",
        "LexicoGradientBoostingClassifier",
    ]
]



# type: ignore
# flake8: noqa
from deepforest import CascadeForestClassifier
from scikit_longitudinal.data_preparation import AggrFunc, SepWav
from scikit_longitudinal.estimators.ensemble.longitudinal_voting.longitudinal_voting import (
    LongitudinalEnsemblingStrategy,
)
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils.multiclass import unique_labels


class AggrFuncMean(AggrFunc):
    def __init__(
        self,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        aggregation_func="mean",
        parallel=False,
        num_cpus=-1,
    ):
        super().__init__(
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            aggregation_func=aggregation_func,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class AggrFuncMedian(AggrFunc):
    def __init__(
        self,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        aggregation_func="median",
        parallel=False,
        num_cpus=-1,
    ):
        super().__init__(
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            aggregation_func=aggregation_func,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class AggrFuncMode(AggrFunc):
    def __init__(
        self,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        aggregation_func="mode",
        parallel=False,
        num_cpus=-1,
    ):
        super().__init__(
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            aggregation_func=aggregation_func,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavMajorityVoting(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.MAJORITY_VOTING,
        stacking_meta_learner=LogisticRegression(),
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavStackingLogisticRegression(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.STACKING,
        stacking_meta_learner=LogisticRegression(),
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavStackingDecisionTree(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.STACKING,
        stacking_meta_learner=DecisionTreeClassifier(),
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavStackingRandomForest(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.STACKING,
        stacking_meta_learner=RandomForestClassifier(),
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavDecayLinearWeighting(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.DECAY_LINEAR_VOTING,
        stacking_meta_learner=None,
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavDecayExponentialWeighting(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.DECAY_EXPONENTIAL_VOTING,
        stacking_meta_learner=None,
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class SepWavDecayCrossValidation(SepWav):
    def __init__(
        self,
        estimator=None,
        features_group=None,
        non_longitudinal_features=None,
        feature_list_names=None,
        voting=LongitudinalEnsemblingStrategy.CV_BASED_VOTING,
        stacking_meta_learner=None,
        n_jobs: int = None,
        parallel: bool = False,
        num_cpus: int = -1,
    ):
        super().__init__(
            estimator=estimator,
            features_group=features_group,
            non_longitudinal_features=non_longitudinal_features,
            feature_list_names=feature_list_names,
            voting=voting,
            stacking_meta_learner=stacking_meta_learner,
            n_jobs=n_jobs,
            parallel=parallel,
            num_cpus=num_cpus,
        )


class CascadeForestClassifier(CascadeForestClassifier):
    def __init__(
        self,
        n_bins=255,
        bin_subsample=200000,
        bin_type="percentile",
        max_layers=20,
        criterion="gini",
        n_estimators=2,
        n_trees=100,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        use_predictor=False,
        predictor="forest",
        predictor_kwargs={},
        backend="custom",
        n_tolerant_rounds=2,
        delta=1e-5,
        partial_mode=False,
        n_jobs=None,
        random_state=None,
        verbose=1,
    ):
        self.classes_ = None
        super().__init__(
            n_bins=n_bins,
            bin_subsample=bin_subsample,
            bin_type=bin_type,
            max_layers=max_layers,
            criterion=criterion,
            n_estimators=n_estimators,
            n_trees=n_trees,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            use_predictor=use_predictor,
            predictor=predictor,
            predictor_kwargs=predictor_kwargs,
            backend=backend,
            n_tolerant_rounds=n_tolerant_rounds,
            delta=delta,
            partial_mode=partial_mode,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
        )

    def fit(self, X, y):
        if self.classes_ is None:
            self.classes_ = unique_labels(y)
        super().fit(X, y)
        print("CascadeForestClassifier is fitted.")
        return self



from typing import Optional, Sequence, Tuple

from sklearn.base import TransformerMixin

from gama.GamaPipeline import GamaPipelineTypeUnion, GamaPipelineType
from gama.postprocessing.base_post_processing import BasePostProcessing
from gama.postprocessing.best_fit import BestFitPostProcessing
from gama.postprocessing.ensemble import EnsemblePostProcessing


class NoPostProcessing(BasePostProcessing):
    """Does nothing, no time will be reserved for post-processing."""

    def to_code(
        self,
        preprocessing: Optional[Sequence[Tuple[str, TransformerMixin]]] = None,
        gama_pipeline_type: Optional[
            GamaPipelineTypeUnion
        ] = GamaPipelineType.ScikitLearn,
    ) -> str:
        raise NotImplementedError(
            "NoPostProcessing has no `to_code` function, since no model is selected."
        )

    def __init__(self, time_fraction: float = 0.0):
        super().__init__(time_fraction)

    def post_process(self, *args, **kwargs) -> None:
        return None


__all__ = ["NoPostProcessing", "BestFitPostProcessing", "EnsemblePostProcessing"]
